from datetime import datetime
from typing import Optional

from pydantic import UUID4, BaseModel

from galileo_core.schemas.core.group_role import GroupRole
from galileo_core.schemas.core.group_visibility import GroupVisibility


class CreateGroupRequest(BaseModel):
    name: str
    description: Optional[str] = None
    visibility: GroupVisibility = GroupVisibility.public


class CreateGroupResponse(CreateGroupRequest):
    id: UUID4
    created_at: datetime


class AddGroupMemberRequest(BaseModel):
    user_id: UUID4
    role: GroupRole = GroupRole.member


class AddGroupMemberResponse(BaseModel):
    user_id: UUID4
    group_role: GroupRole



from enum import Enum
from typing import List, Literal, Union

from pydantic import BaseModel, Field
from pydantic.types import UUID4
from typing_extensions import Annotated


class ContentPartType(str, Enum):
    text = "text"
    file = "file"


class TextContentPart(BaseModel):
    """A text segment within a message."""

    type: Literal[ContentPartType.text] = ContentPartType.text
    text: str


class FileContentPart(BaseModel):
    """Reference to a file associated with this message.

    The file_id can be resolved via the ``files`` dict returned on
    trace/span detail responses, which contains metadata such as
    modality, MIME type, and a presigned download URL.
    """

    type: Literal[ContentPartType.file] = ContentPartType.file
    file_id: UUID4


ContentPart = Annotated[
    Union[TextContentPart, FileContentPart],
    Field(discriminator="type"),
]

MessageContent = Union[str, List[ContentPart]]



from enum import Enum


class LLMIntegration(str, Enum):
    anthropic = "anthropic"
    azure = "azure"
    aws_bedrock = "aws_bedrock"
    aws_sagemaker = "aws_sagemaker"
    databricks = "databricks"
    vertex_ai = "vertex_ai"
    openai = "openai"
    writer = "writer"
    mistral = "mistral"
    nvidia = "nvidia"



from pydantic import UUID4, BaseModel

from galileo_core.schemas.core.collaboration_role import CollaboratorRole


class CollaboratorCreateBase(BaseModel):
    role: CollaboratorRole = CollaboratorRole.viewer


class UserCollaboratorCreate(CollaboratorCreateBase):
    user_id: UUID4


class GroupCollaboratorCreate(CollaboratorCreateBase):
    group_id: UUID4



from enum import Enum
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class CustomizedScorerName(str, Enum):
    action_advancement = "_customized_agentic_workflow_success"
    action_completion = "_customized_agentic_session_success"
    chunk_attribution_utilization_plus = "_customized_chunk_attribution_utilization_gpt"
    completeness_plus = "_customized_completeness_gpt"
    context_adherence_plus = "_customized_groundedness"
    correctness = "_customized_factuality"
    ground_truth_adherence = "_customized_ground_truth_adherence"
    instruction_adherence = "_customized_instruction_adherence"
    prompt_injection_plus = "_customized_prompt_injection_gpt"
    tool_errors = "_customized_tool_error_rate"
    tool_selection_quality = "_customized_tool_selection_quality"
    sexist_plus = "_customized_sexist_gpt"
    input_sexist_plus = "_customized_input_sexist_gpt"
    toxicity_plus = "_customized_toxicity_gpt"
    input_toxicity_plus = "_customized_input_toxicity_gpt"


class CustomizedScorer(BaseModel):
    scorer_name: CustomizedScorerName = Field(..., description="Name of the customized scorer.")
    model_alias: Optional[str] = Field(default=None, description="Model alias to use for scoring.")
    num_judges: Optional[int] = Field(default=None, ge=1, le=10, description="Number of judges for the scorer.")

    model_config = ConfigDict(
        # Avoid Pydantic's protected namespace warning since we want to use
        # `model_alias` as a field name.
        protected_namespaces=(),
    )



from enum import Enum
from random import choice
from typing import List, Literal, Sequence, Union

from pydantic import BaseModel, Field
from typing_extensions import Annotated

from galileo_core.schemas.protect.subscription_config import SubscriptionConfig


class ActionType(str, Enum):
    OVERRIDE = "OVERRIDE"
    PASSTHROUGH = "PASSTHROUGH"


class ActionResult(BaseModel):
    type: ActionType = Field(description="Type of action that was taken.")
    value: str = Field(description="Value of the action that was taken.")


class BaseAction(BaseModel):
    type: ActionType = Field(description="Type of action to take.")
    subscriptions: List[SubscriptionConfig] = Field(
        default_factory=list,
        description="List of subscriptions to send a notification to when this action is applied and the ruleset status matches any of the configured statuses.",
    )

    def apply(self, response: str) -> ActionResult:
        raise NotImplementedError


class OverrideAction(BaseAction):
    type: Literal[ActionType.OVERRIDE] = ActionType.OVERRIDE
    choices: Sequence[str] = Field(
        description="List of choices to override the response with. If there are multiple choices, one will be chosen at random when applying this action.",
        min_length=1,
    )

    def apply(self, response: str) -> ActionResult:
        return ActionResult(type=self.type, value=choice(self.choices))


class PassthroughAction(BaseAction):
    type: Literal[ActionType.PASSTHROUGH] = ActionType.PASSTHROUGH

    def apply(self, response: str) -> ActionResult:
        return ActionResult(type=self.type, value=response)


Action = Annotated[Union[OverrideAction, PassthroughAction], Field(discriminator="type")]



from os import environ
from ssl import SSLContext, create_default_context
from typing import Optional, Union
from warnings import warn

from galileo_core.helpers.logger import logger


def get_ssl_context(
    ssl_context: Union[SSLContext, bool] = True, logging_extra: Optional[dict] = None
) -> Union[SSLContext, bool]:
    """
    Parse an SSL context.

    We allow for the SSL context to be set in a few different ways:
    1. If the user provides an SSLContext object, we use that.
    2. If the user provides `False`, we disable SSL.
    3. If the user provides `True`, we create a default SSL context.
    4. If the user sets `SSL_CERT_FILE` or `SSL_CERT_DIR`, we use those to create a custom SSL context.

    Inspiration: https://www.python-httpx.org/advanced/ssl/#client-side-certificates
    """
    ssl_cert_file = environ.get("SSL_CERT_FILE")
    ssl_cert_dir = environ.get("SSL_CERT_DIR")
    if ssl_context is False:
        warn("SSL is disabled. This is not recommended for production use.", category=UserWarning)
        logger.debug("SSL validation is disabled.", extra=logging_extra)
    elif ssl_cert_file or ssl_cert_dir:
        logger.debug(
            f"SSL context is set to custom cert file ({ssl_cert_file}) or directory ({ssl_cert_dir}).",
            extra=logging_extra,
        )
        ssl_context = create_default_context(cafile=ssl_cert_file, capath=ssl_cert_dir)
    else:
        logger.debug("SSL context is set to the default context.", extra=logging_extra)
        ssl_context = True
    return ssl_context



from datetime import datetime
from enum import Enum
from typing import List, Literal, Optional, Union

from pydantic import UUID4, BaseModel, Field, ValidationInfo, field_validator
from typing_extensions import Annotated


class Operator(str, Enum):
    eq = "eq"
    ne = "ne"
    contains = "contains"
    one_of = "one_of"
    not_in = "not_in"
    gt = "gt"
    gte = "gte"
    lt = "lt"
    lte = "lte"
    between = "between"


class FilterType(str, Enum):
    number = "number"
    string = "string"
    enum = "enum"
    collection = "collection"
    map = "map"
    custom_number = "custom_number"
    custom_boolean = "custom_boolean"
    custom_uuid = "custom_uuid"
    id = "id"
    date = "date"
    between = "between"


class FilterBase(BaseModel):
    name: Optional[str]


StringOperator = Literal[Operator.eq, Operator.ne, Operator.contains]


class StringFilter(FilterBase):
    """
    Filters on a string field.
    """

    filter_type: Literal[FilterType.string] = FilterType.string
    value: str
    operator: StringOperator
    case_sensitive: bool = True


CollectionOperator = Literal[Operator.contains, Operator.not_in]


class CollectionFilter(FilterBase):
    """
    Filters for string items in a collection/list.
    """

    filter_type: Literal[FilterType.collection] = FilterType.collection
    value: str
    operator: CollectionOperator


MapOperator = Literal[Operator.one_of, Operator.not_in, Operator.eq, Operator.ne]


class MapFilter(FilterBase):
    """
    Filters for string items in a map / dictionary.
    """

    filter_type: Literal[FilterType.map] = FilterType.map
    operator: MapOperator
    key: str
    value: Union[str, List[str]]

    @field_validator("value", mode="before")
    def validate_value(cls, value: Union[str, List[str]], info: ValidationInfo) -> Union[str, List[str]]:
        operator = info.data.get("operator")
        if operator in (Operator.one_of, Operator.not_in):
            if not isinstance(value, list):
                raise ValueError(f"Value must be a list for operator {operator}.")
        elif operator in (Operator.eq, Operator.ne):
            if isinstance(value, list):
                raise ValueError(f"Value must be a string for operator {operator}.")
        return value


EnumOperator = Literal[Operator.eq, Operator.ne]


class EnumFilter(FilterBase):
    """
    Filters on a string field, with limited categories.
    """

    filter_type: Literal[FilterType.enum] = FilterType.enum
    value: str
    operator: EnumOperator


class IDFilter(FilterBase):
    """
    Filters on a UUID field.
    """

    filter_type: Literal[FilterType.id] = FilterType.id
    value: UUID4


DateOperator = Literal[Operator.eq, Operator.ne, Operator.gt, Operator.gte, Operator.lt, Operator.lte]


class DateFilter(FilterBase):
    """
    Filters on a datetime field.
    """

    filter_type: Literal[FilterType.date] = FilterType.date
    value: datetime
    operator: DateOperator


NumberOperator = Literal[
    Operator.eq, Operator.ne, Operator.gt, Operator.gte, Operator.lt, Operator.lte, Operator.between
]


class CustomNumberFilter(FilterBase):
    filter_type: Literal[FilterType.custom_number] = FilterType.custom_number
    value: Union[int, List[int]]
    operator: NumberOperator


class CustomBooleanFilter(FilterBase):
    filter_type: Literal[FilterType.custom_boolean] = FilterType.custom_boolean
    value: bool


class CustomUUIDFilter(FilterBase):
    filter_type: Literal[FilterType.custom_uuid] = FilterType.custom_uuid
    value: UUID4


QueryFilterV2 = Annotated[
    Union[
        CollectionFilter,
        CustomBooleanFilter,
        CustomNumberFilter,
        CustomUUIDFilter,
        DateFilter,
        EnumFilter,
        IDFilter,
        MapFilter,
        StringFilter,
    ],
    Field(discriminator="filter_type"),
]



from contextlib import contextmanager
from ssl import SSLContext
from threading import local
from typing import Any, Dict, Iterator, Optional, Union
from urllib.parse import urljoin

from httpx import AsyncClient, Client, HTTPError, Response, Timeout
from pydantic import BaseModel, ConfigDict, Field, HttpUrl, SecretStr

from galileo_core.constants.http_headers import HttpHeaders
from galileo_core.constants.request_method import RequestMethod
from galileo_core.exceptions.http import GalileoHTTPException
from galileo_core.helpers.execution import async_run
from galileo_core.helpers.logger import logger

DEFAULT_TIMEOUT_SECONDS = 60.0


class ApiClient(BaseModel):
    host: HttpUrl
    jwt_token: SecretStr
    # Used by the openapi generated resources
    raise_on_unexpected_status: bool = False
    ssl_context: Union[SSLContext, bool] = True
    # Thread-local storage for AsyncClient instances
    thread_local: local = Field(default_factory=local, exclude=True)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def async_client(self) -> AsyncClient:
        """Get or create an AsyncClient for the current thread."""
        if not hasattr(self.thread_local, "client"):
            self.thread_local.client = AsyncClient(
                base_url=self.host.unicode_string(),
                verify=self.ssl_context,
                timeout=Timeout(DEFAULT_TIMEOUT_SECONDS, connect=5.0),
            )
        return self.thread_local.client

    @property
    def sync_client(self) -> Client:
        """Get or create a synchronous Client for the current thread."""
        if not hasattr(self.thread_local, "sync_client"):
            self.thread_local.sync_client = Client(
                base_url=self.host.unicode_string(),
                verify=self.ssl_context,
                timeout=Timeout(DEFAULT_TIMEOUT_SECONDS, connect=5.0),
            )
        return self.thread_local.sync_client

    @property
    def auth_header(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.jwt_token.get_secret_value()}"}

    @staticmethod
    def validate_response(response: Response, raise_on_error: bool = True) -> None:
        for header, value in response.headers.items():
            # Log all Galileo headers. These are headers that start with `Galileo-Request-`.
            # These can be in any case, so we lower-case them for comparison.
            if header.lower().startswith("galileo-request-"):
                logger.debug(f"{header.title()}: {value}.")
        if raise_on_error:
            try:
                response.raise_for_status()
            except HTTPError:
                raise GalileoHTTPException(
                    f"Galileo API returned HTTP status code {response.status_code}. Error was: {response.text}",
                    response.status_code,
                    response.text,
                )

    @staticmethod
    async def make_request(
        request_method: RequestMethod,
        base_url: str,
        endpoint: str,
        ssl_context: Union[SSLContext, bool] = True,
        read_timeout: float = DEFAULT_TIMEOUT_SECONDS,
        return_raw_response: bool = False,
        async_client: Optional[AsyncClient] = None,
        **kwargs: Any,
    ) -> Any:
        url = urljoin(base_url, endpoint)
        logger.debug(f"Making request to {url}.")
        local_client = False
        if async_client is None:
            # Create a new AsyncClient instance.
            logger.debug("Creating new AsyncClient instance for this request.")
            async_client = AsyncClient(
                base_url=base_url, timeout=Timeout(read_timeout, connect=5.0), verify=ssl_context
            )
            local_client = True

        response = await async_client.request(
            method=request_method.value,
            url=url,
            timeout=Timeout(read_timeout, connect=5.0),
            **kwargs,
        )
        # Skip validation when return_raw_response=True, as the caller will handle the response
        ApiClient.validate_response(response, raise_on_error=not return_raw_response)
        logger.debug(f"Response was received from {url}.")

        if local_client:
            await async_client.aclose()
            logger.debug("Closed AsyncClient instance.")

        # If the caller asked for the raw Response object, just return it.
        if return_raw_response:
            return response

        # Some successful responses (e.g. HTTP 204 No Content) legitimately have no body.
        # Attempting to call `response.json()` in these cases raises a JSONDecodeError.
        # We treat 204 responses or any response with an empty body as having no payload
        # and return `None` to the caller.
        if response.status_code == 204 or not response.content:
            return None

        # Otherwise attempt to parse the response as JSON. If this fails, fall back to
        # returning the raw text content so the caller still receives the payload.
        try:
            return response.json()
        except ValueError:
            return response.text

    @contextmanager
    def stream_request(
        self,
        method: RequestMethod,
        path: str,
        content_headers: Dict[str, str] = HttpHeaders.json(),
        read_timeout: float = DEFAULT_TIMEOUT_SECONDS,
        **kwargs: Any,
    ) -> Iterator[Response]:
        kwargs.pop("return_raw_response", None)
        headers = {**content_headers, **self.auth_header}
        url = urljoin(self.host.unicode_string(), path)
        logger.debug(f"Streaming request to {url}.")

        with self.sync_client.stream(
            method=method.value,
            url=url,
            headers=headers,
            timeout=Timeout(read_timeout, connect=5.0),
            **kwargs,
        ) as response:
            ApiClient.validate_response(response)
            yield response

    def request(
        self,
        method: RequestMethod,
        path: str,
        content_headers: Dict[str, str] = HttpHeaders.json(),
        **kwargs: Any,
    ) -> Any:
        return async_run(self.arequest(method=method, path=path, content_headers=content_headers, **kwargs))

    async def arequest(
        self,
        method: RequestMethod,
        path: str,
        content_headers: Dict[str, str] = HttpHeaders.json(),
        **kwargs: Any,
    ) -> Any:
        return await ApiClient.make_request(
            request_method=method,
            base_url=self.host.unicode_string(),
            endpoint=path,
            headers={**content_headers, **self.auth_header},
            ssl_context=self.ssl_context,
            async_client=self.async_client,
            **kwargs,
        )



from typing import Any, Callable, Dict, List

from galileo_core.helpers.logger import logger
from galileo_core.schemas.core.pagination import PaginationDefaults, PaginationRequest, PaginationResponse


def paginated_request(
    request_partial: Callable[..., Dict],
    result_key: str,
    starting_token: int = PaginationDefaults.starting_token,
    limit: int = PaginationDefaults.limit,
) -> List[Any]:
    """
    Get results from a request that may be paginated.

    We make an API request (from a partial function) and get the results from the response. If the response is paginated,
    we make a follow-up request with the next starting token and append the results to the list.

    Parameters
    ----------
    request_partial : Callable[..., Dict]
        Partial function for making an API request.
    result_key : str
        Key in the response JSON to get the results from.
    starting_token : int, optional
        Starting token for the request, by default 0
    limit : int, optional
        Limit for the number of items to request, by default 25

    Returns
    -------
    List[Any]
        List of results from the API request.
    """
    logger.debug(f"Requesting {limit} items starting from {starting_token}...")
    response = request_partial(params=PaginationRequest(starting_token=starting_token, limit=limit).model_dump())
    results = response.get(result_key, list())
    # If the response is paginated, make a follow-up request with the next starting token.
    pagination_response = PaginationResponse.model_validate(response)
    if pagination_response.paginated and pagination_response.next_starting_token is not None:
        # Recursively get the results from the next page.
        results += paginated_request(
            request_partial, result_key, starting_token=pagination_response.next_starting_token, limit=limit
        )
    return results



from typing import Literal

from pydantic import BaseModel

from galileo_core.schemas.core.integration.integration_type import LLMIntegration


class WriterExtras(BaseModel):
    organization_id: str


class WriterIntegrationCreate(WriterExtras):
    name: Literal[LLMIntegration.writer] = LLMIntegration.writer
    token: str



from functools import partial
from typing import List, Optional
from urllib.parse import parse_qs, urlparse

from pydantic import UUID4

from galileo_core.constants.request_method import RequestMethod
from galileo_core.constants.routes import Routes
from galileo_core.helpers.logger import logger
from galileo_core.helpers.pagination import paginated_request
from galileo_core.schemas.base_config import GalileoConfig
from galileo_core.schemas.core.auth_method import AuthMethod
from galileo_core.schemas.core.user import (
    CreateUserRequest,
    CreateUserResponse,
    InviteUsersRequest,
    UpdateUserRoleRequest,
    User,
)
from galileo_core.schemas.core.user_role import UserRole


def get_sign_up_token(
    email: str,
    config: Optional[GalileoConfig] = None,
) -> str:
    """
    Get an access token from a signup link for a new user.

    Parameters
    ----------
    email: str
        The user's email address.

    Returns
    -------
    str
        A JWT access token as a string.

    Raises
    ------
    ValueError
        If authentication fails or the token cannot be retrieved.
    """
    config = config or GalileoConfig.get()

    response = config.api_client.request(
        RequestMethod.POST,
        Routes.signup_link,
        params={"user_email": email},
    )

    parsed_url = urlparse(response["signup_url"])
    query_params = parse_qs(parsed_url.query)
    token = query_params.get("token", [None])[0]

    if not token:
        raise ValueError("Token not found in URL.")
    return token


def create_user(
    email: str,
    password: str,
    auth_method: AuthMethod = AuthMethod.email,
    role: UserRole = UserRole.read_only,
    config: Optional[GalileoConfig] = None,
) -> CreateUserResponse:
    """
    Create user.

    Parameters
    ----------
    email: str
        The user's email address.

    password: str
        The user's password.

    auth_method: AuthMethod = AuthMethod.email
        The method of authentication for the user.

    role: UserRole = UserRole.read_only
        The user's assigned role.


    Returns
    -------
    CreateUserResponse
        Created user.
    """
    config = config or GalileoConfig.get()
    logger.debug("Creating a new user...")

    new_user = CreateUserRequest(
        email=email,
        auth_method=auth_method,
        password=password,
        role=role,
    )
    sign_up_token = get_sign_up_token(email=email, config=config)

    response_dict = config.api_client.request(
        RequestMethod.POST, Routes.users, json=new_user.model_dump(mode="json"), params={"signup_token": sign_up_token}
    )
    user = CreateUserResponse.model_validate(response_dict)
    logger.debug(f"Created user {str(user.id)} with role {user.role.value}.")
    return user


def get_current_user(config: Optional[GalileoConfig] = None) -> User:
    """
    Get the current user.

    Returns
    -------
    User
        Current user.
    """
    config = config or GalileoConfig.get()
    logger.debug("Getting current user...")
    response_dict = config.api_client.request(RequestMethod.GET, Routes.current_user)
    user = User.model_validate(response_dict)
    logger.debug(f"Got current user {user.email}.")
    return user


def invite_users(
    emails: List[str],
    role: UserRole = UserRole.user,
    group_ids: Optional[List[UUID4]] = None,
    auth_method: AuthMethod = AuthMethod.email,
    config: Optional[GalileoConfig] = None,
) -> None:
    """
    Invite users.

    Parameters
    ----------
    emails : List[str]
        List of emails to invite.
    role : UserRole, optional
        Roles to grant invited users, by default UserRole.user
    group_ids : Optional[List[UUID4]], optional
        Group IDs to add the users to, by default None, which means they are not added to any group.
    auth_method : AuthMethod, optional
        Authentication method to use, by default AuthMethod.email
    """
    config = config or GalileoConfig.get()
    group_ids = group_ids or list()
    request = InviteUsersRequest(emails=emails, role=role, group_ids=group_ids, auth_method=auth_method)
    logger.debug(f"Inviting users {request.emails} with role {request.role}...")
    config.api_client.request(RequestMethod.POST, Routes.invite_users, json=request.model_dump(mode="json"))
    logger.debug(f"Invited users {request.emails} with role {request.role}.")


def list_users(config: Optional[GalileoConfig] = None) -> List[User]:
    """
    List all users.

    Returns
    -------
    List[User]
        List of all users.
    """
    config = config or GalileoConfig.get()
    logger.debug("Listing users...")
    all_users = paginated_request(partial(config.api_client.request, RequestMethod.GET, Routes.users), "users")
    users = [User.model_validate(user) for user in all_users]
    logger.debug(f"Listed all users, found {len(users)} users.")
    return users


def update_user(user_id: UUID4, role: UserRole, config: Optional[GalileoConfig] = None) -> User:
    """
    Update user.

    Parameters
    ----------
    user_id : User.id
        User ID to update.
    role : UserRole
        New role to assign to the user.

    Returns
    -------
    User
        Updated user.
    """
    config = config or GalileoConfig.get()
    logger.debug(f"Updating user {str(user_id)} to {role}...")

    request = UpdateUserRoleRequest(role=role)
    response_dict = config.api_client.request(
        RequestMethod.PUT, Routes.update_user.format(user_id=user_id), json=request.model_dump(mode="json")
    )
    user = User.model_validate(response_dict)
    logger.debug(f"Updated user {str(user_id)} role to {user.role.value}.")
    return user



"""
EventLoopThreadPool provides a mechanism for running asyncio coroutines from synchronous code.

This module contains implementations for:
1. EventLoopThreadInterface - An abstract interface for event loop thread implementations
2. EventLoopThread - A single thread with its own event loop for executing coroutines
3. EventLoopThreadPool - A pool of EventLoopThread instances that are initialized lazily

The EventLoopThreadPool is particularly useful in scenarios where:
- You need to execute async code from a synchronous context
- You want to parallelize async operations across multiple event loops
- You need to maintain a pool of reusable event loops for performance reasons

Examples:
    Basic usage with waiting for results:
    ```python
    # Create a thread pool
    pool = EventLoopThreadPool(name="MyThreadPool", num_threads=4)

    # Define an async function
    async def fetch_data(url):
        # ... async code here
        return result

    # Execute the async function and wait for results
    result = pool.submit(lambda: fetch_data("https://example.com"))
    ```

    Fire-and-forget usage:
    ```python
    # Submit without waiting for result
    future = pool.submit(lambda: process_data_async(data), wait_for_result=False)

    # Optionally check the result later
    result = future.result()  # This will block until complete
    ```

    Direct coroutine submission:
    ```python
    # Create and submit a coroutine directly
    coroutine = fetch_data("https://example.com")
    result = pool.submit(coroutine)
    ```

The module handles gevent integration automatically and will adapt its behavior
when running in a gevent-patched environment.
"""

import asyncio
import os
import random
import threading
from abc import abstractmethod
from asyncio import AbstractEventLoop
from concurrent.futures import Future
from logging import getLogger
from typing import Any, Callable, Coroutine, List, Union

logger = getLogger(__name__)


# EventLoopThread doesn't play nice with gevent because gevent patches threading.
# So if we are in gevent patched environment, we have different control flow.
_GEVENT_PATCHED = False
try:
    import gevent  # type: ignore[import-not-found]
    import gevent.monkey  # type: ignore[import-not-found]

    # Check if threading specifically is patched
    _GEVENT_PATCHED = gevent.monkey.is_module_patched("threading")
except ImportError:
    pass  # gevent not installed or not used


class EventLoopThreadInterface:
    def submit(
        self,
        async_fn: Union[Callable[[], Coroutine[Any, Any, Any]], Coroutine[Any, Any, Any]],
        wait_for_result: bool = True,
    ) -> Any:
        """
        Submit an async function or coroutine.

        If `wait_for_result` is True, this method will block until the async function
        completes and returns its result (or raises an exception). Otherwise, it returns
        a concurrent.futures.Future immediately.
        """
        pass

    @abstractmethod
    def stop(self, timeout_sec: int = 2) -> None:
        """
        For testing only.
        :param timeout_sec:
        :return:
        """
        pass


class EventLoopThread(EventLoopThreadInterface, threading.Thread):
    """
    A thread that runs its own event loop so you can submit coroutines to be executed
    on it and wait for results from synchronous code.
    """

    def __init__(self, name: str = "EventLoopThread", daemon: bool = True) -> None:
        super().__init__(name=name)
        self.daemon = daemon
        self.loop: AbstractEventLoop
        if not _GEVENT_PATCHED:
            try:
                import uvloop  # type: ignore[import-not-found]

                self.loop = uvloop.new_event_loop()
            except ImportError:
                self.loop = asyncio.new_event_loop()
        else:
            self.loop = asyncio.new_event_loop()

    def run(self) -> None:
        # Set the event loop for this thread.
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    def submit(
        self,
        async_fn: Union[Callable[[], Coroutine[Any, Any, Any]], Coroutine[Any, Any, Any]],
        wait_for_result: bool = True,
    ) -> Any:
        coro: Coroutine[Any, Any, Any] = async_fn if isinstance(async_fn, Coroutine) else async_fn()
        future: Future[Any] = asyncio.run_coroutine_threadsafe(coro, self.loop)

        if wait_for_result:
            # Block until the result is ready (or an exception is raised).
            result = future.result()
            return result
        else:
            return future

    def stop(self, timeout_sec: int = 2) -> None:
        """Stop the event loop."""
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.join(timeout=timeout_sec)


class EventLoopThreadPool(EventLoopThreadInterface):
    """
    Manage collection of EventLoopThread instances and also initailize them lazily.

    Lazy initialization is important when running in forked processes.
    """

    def __init__(self, name: str = "EventLoopThreadPool", num_threads: int = 16) -> None:
        self.name = name
        self.num_threads = num_threads
        self.threads: List[EventLoopThread] = []
        self.init_pid = os.getpid()  # Store the PID when the pool is initialized

        self.init_lock = threading.Lock()

    def submit(
        self,
        async_fn: Union[Callable[[], Coroutine[Any, Any, Any]], Coroutine[Any, Any, Any]],
        wait_for_result: bool = True,
    ) -> Any:
        if _GEVENT_PATCHED:
            import gevent
            from gevent.event import AsyncResult  # type: ignore[import-not-found]

            async_result: AsyncResult[Any] = AsyncResult()

            def run_coro_in_temp_loop() -> None:
                try:
                    coro = async_fn if asyncio.iscoroutine(async_fn) else async_fn()
                    result: Any = asyncio.run(coro)  # type: ignore[arg-type]
                    async_result.set(result)
                except Exception as e:
                    async_result.set_exception(e)

            gevent.spawn(run_coro_in_temp_loop)

            if wait_for_result:
                return async_result.get()
            return async_result
        else:
            # If threads list is empty or we're in a different process (forked), initialize new threads
            if len(self.threads) == 0 or os.getpid() != self.init_pid:
                with self.init_lock:
                    if len(self.threads) == 0 or os.getpid() != self.init_pid:
                        self.threads = [EventLoopThread(f"{self.name}-{i}") for i in range(self.num_threads)]
                        self.init_pid = os.getpid()  # Update the PID
                        for thread in self.threads:
                            thread.start()
                        logger.debug(f"Initialized EventLoopThreadPool {self.name} in PID {self.init_pid}")
            return random.choice(self.threads).submit(async_fn, wait_for_result)

    def stop(self, timeout_sec: int = 2) -> None:
        if self.threads:
            for thread in self.threads:
                thread.stop(timeout_sec)



from enum import Enum
from typing import List, Literal, Optional, Union

from pydantic import BaseModel, Field, StringConstraints
from typing_extensions import Annotated

FeedbackTagType = Annotated[str, StringConstraints(min_length=1, max_length=255, strip_whitespace=True)]


class FeedbackType(str, Enum):
    like_dislike = "like_dislike"
    star = "star"
    score = "score"
    tags = "tags"
    text = "text"


class AnnotationType(str, Enum):
    like_dislike = "like_dislike"
    star = "star"
    score = "score"
    tags = "tags"
    text = "text"


class FeedbackRatingInfo(BaseModel):
    feedback_type: FeedbackType
    value: Union[bool, int, str, List[FeedbackTagType]]
    explanation: Optional[str]


class AnnotationRatingInfo(BaseModel):
    annotation_type: AnnotationType
    value: Union[bool, int, str, List[FeedbackTagType]]
    explanation: Optional[str]


class AnnotationLikeDislikeAggregate(BaseModel):
    annotation_type: Literal[AnnotationType.like_dislike] = AnnotationType.like_dislike
    like_count: int
    dislike_count: int
    unrated_count: int


class AnnotationStarAggregate(BaseModel):
    annotation_type: Literal[AnnotationType.star] = AnnotationType.star
    average: float
    counts: dict[int, int]
    unrated_count: int


class AnnotationScoreAggregate(BaseModel):
    annotation_type: Literal[AnnotationType.score] = AnnotationType.score
    average: float
    unrated_count: int


class AnnotationTagsAggregate(BaseModel):
    annotation_type: Literal[AnnotationType.tags] = AnnotationType.tags
    counts: dict[str, int]
    unrated_count: int


class AnnotationTextAggregate(BaseModel):
    annotation_type: Literal[AnnotationType.text] = AnnotationType.text
    count: int
    unrated_count: int


class AnnotationAggregate(BaseModel):
    aggregate: Annotated[
        AnnotationLikeDislikeAggregate
        | AnnotationStarAggregate
        | AnnotationScoreAggregate
        | AnnotationTagsAggregate
        | AnnotationTextAggregate,
        Field(discriminator="annotation_type"),
    ]



from typing import Dict, List, Union

SingleMetricValue = Union[float, int, str, None]
MetricValueType = Union[SingleMetricValue, List[SingleMetricValue], Dict[str, SingleMetricValue]]
UserMetricType = Union[float, int, bool, str, None, dict, list]



from typing import Optional

from galileo_core.exceptions.base import BaseGalileoException


class ExecutionError(BaseGalileoException):
    """Raised when there is an issue with the execution of a task or process."""

    error_type: Optional[str] = None


class MetricNotFoundError(ExecutionError):
    """Raised when attempting to access a metric that doesn't exist in the Metrics object."""

    def __init__(self, metric_name: str, extra: Optional[dict] = None):
        message = (
            f"Metric '{metric_name}' does not exist. "
            f"The metric you have attempted to retrieve is either not in the required metrics, "
            f"or you have used an incorrect name."
        )
        super().__init__(message, extra)
        self.metric_name = metric_name


class MetricStatusError(ExecutionError):
    """Base class for errors when a metric has a non-success status."""

    def __init__(self, metric_name: str, status: str, extra: Optional[dict] = None):
        message = f"Metric '{metric_name}' has status '{status}'."
        super().__init__(message, extra)
        self.metric_name = metric_name
        self.status = status


class MetricFailedError(MetricStatusError):
    """Raised when attempting to access a metric that has 'failed' status."""

    def __init__(self, metric_name: str, extra: Optional[dict] = None):
        super().__init__(metric_name, "failed", extra)


class MetricErrorError(MetricStatusError):
    """Raised when attempting to access a metric that has 'error' status."""

    def __init__(self, metric_name: str, extra: Optional[dict] = None):
        super().__init__(metric_name, "error", extra)


class MetricNotComputedError(MetricStatusError):
    """Raised when attempting to access a metric that has 'not_computed' status."""

    def __init__(self, metric_name: str, extra: Optional[dict] = None):
        super().__init__(metric_name, "not_computed", extra)


class MetricNotApplicableError(MetricStatusError):
    """Raised when attempting to access a metric that has 'not_applicable' status."""

    def __init__(self, metric_name: str, extra: Optional[dict] = None):
        super().__init__(metric_name, "not_applicable", extra)


class MetricPendingError(MetricStatusError):
    """Raised when attempting to access a metric that has 'pending' status (still queued)."""

    def __init__(self, metric_name: str, extra: Optional[dict] = None):
        super().__init__(metric_name, "pending", extra)


class MetricComputingError(MetricStatusError):
    """Raised when attempting to access a metric that has 'computing' status (still in progress)."""

    def __init__(self, metric_name: str, extra: Optional[dict] = None):
        super().__init__(metric_name, "computing", extra)



from typing import List, Union

from pydantic import BaseModel, ConfigDict

from galileo_core.schemas.shared.content_parts import FileContentPart, MessageContent, TextContentPart
from galileo_core.schemas.shared.message_role import MessageRole


class Message(BaseModel):
    content: MessageContent
    role: Union[str, MessageRole]

    model_config = ConfigDict(extra="allow")

    @property
    def message(self) -> str:
        role = self.role.value if isinstance(self.role, MessageRole) else self.role
        if isinstance(self.content, str):
            return f"{role}: {self.content}"
        parts: List[str] = []
        for part in self.content:
            if isinstance(part, TextContentPart):
                parts.append(part.text)
            elif isinstance(part, FileContentPart):
                parts.append(f"[file:{part.file_id}]")
        return f"{role}: {''.join(parts)}"
